/* -------------- scheduler management interface functions -------------- */

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/resource.h>

int bossa_connect(int parent, int child, int argc, int *argv);
int bossa_attach_info(int parent);
int bossa_print(char *s, int sz);

#define BOSSA_ID_NUM 0xCE
#define IOCTL_CONNECT        _IOWR(BOSSA_ID_NUM, 0, int)
#define IOCTL_PRINT          _IOR(BOSSA_ID_NUM, 1, int)
#define IOCTL_ATTACH_INFO    _IOWR(BOSSA_ID_NUM, 2, int)
#define IOCTL_SCHED	     _IOWR(BOSSA_ID_NUM, 3, int)
#define IOCTL_NBSCHED        _IOWR(BOSSA_ID_NUM, 4, int)
#define IOCTL_NAMES          _IOR(BOSSA_ID_NUM, 5, int)
#define IOCTL_DISCONNECT     _IOWR(BOSSA_ID_NUM, 6, int)
#define IOCTL_CONNECT_ROOT   _IOWR(BOSSA_ID_NUM, 7, int)
#define BOSSA_DEVICE_NAME "bossa"

struct dev_bossa_attach_info_struct {
  int parent;
  int parent_attach_argc;
  char parent_attach_argv[300];
};

struct dev_bossa_attach_info_struct attach_info_data;

struct dev_bossa_connect_struct {
  int parent;
  int child;
  int data[10];
};

struct dev_bossa_connect_root_struct {
  int new_root;
  int data[10];
};

struct dev_bossa_disconnect_struct {
  int tgt;
};

struct dev_bossa_print_struct {
  char *s;
  int sz;
};

int bossa_attach_info(int parent) {
  attach_info_data.parent = parent;
  {
    int fd, res;
    fd = open("/dev/bossa/" BOSSA_DEVICE_NAME, O_RDWR);
    res = ioctl(fd,IOCTL_ATTACH_INFO,(char *)&attach_info_data);
    if (res) perror("bossa_attach_info");
    close(fd);
    return res;
  }
}

int bossa_connect(int parent, int child, int argc, int *argv) {
  struct dev_bossa_connect_struct _data;
  int i;
  _data.parent = parent;
  _data.child = child;
  for(i=0; i!=argc; i++)
    _data.data[i] = argv[i];
  {
    int fd, res;
    fd = open("/dev/bossa/" BOSSA_DEVICE_NAME, O_RDWR);
    res = ioctl(fd,IOCTL_CONNECT,(char *)&_data);
    if (res) perror("bossa_connect");
    close(fd);
    return res;
  }
}

int bossa_connect_root(int new_root, int argc, int *argv) {
  struct dev_bossa_connect_root_struct _data;
  int i;
  _data.new_root = new_root;
  for(i=0; i!=argc; i++)
    _data.data[i] = argv[i];
  {
    int fd, res;
    fd = open("/dev/bossa/" BOSSA_DEVICE_NAME, O_RDWR);
    res = ioctl(fd,IOCTL_CONNECT_ROOT,(char *)&_data);
    if (res) perror("bossa_connect_root");
    close(fd);
    return res;
  }
}

int bossa_disconnect(int tgt) {
  struct dev_bossa_disconnect_struct _data;
  int i;
  _data.tgt = tgt;
  {
    int fd, res;
    fd = open("/dev/bossa/" BOSSA_DEVICE_NAME, O_RDWR);
    res = ioctl(fd,IOCTL_DISCONNECT,(char *)&_data);
    if (res) perror("bossa_disconnect");
    close(fd);
    return res;
  }
}

int bossa_print(char *s, int sz) {
  struct dev_bossa_print_struct _data;
  _data.s = s;
  _data.sz = sz;
  {
    int fd, res;
    fd = open("/dev/bossa/" BOSSA_DEVICE_NAME, O_RDONLY);
    res = ioctl(fd,IOCTL_PRINT,(char *)&_data);
    if (res) perror("bossa_print");
    close(fd);
    return res;
  }
}

int scheduler_names(char *s, int sz) {
  struct dev_bossa_print_struct _data;
  _data.s = s;
  _data.sz = sz;
  {
    int fd, res;
    fd = open("/dev/bossa/" BOSSA_DEVICE_NAME, O_RDONLY);
    res = ioctl(fd,IOCTL_NAMES,(char *)&_data);
    if (res) perror("bossa_names");
    close(fd);
    return res;
  }
}

void load_module(char *s) {
  int fd;
  char tmp[100];
  tmp[0] = '\0';
  strcat(tmp,"/dev/bossa/bossa_");
  strcat(tmp,s);
  fd = open(tmp, O_RDONLY);
  close(fd);
}

int find_scheduler(char *b, char *s) {
  int i;
  int len = strlen(s);
  for(i=0; i!=20; i++) {
    if (!(strcmp(s,b+(i*25)))) return i;
  }
  printf("scheduler %s not found\n",s);
  exit(0);
}  

void mount(char *parent_name, char *child_name,
	  int child_property_count,
	  int *child_properties) {
  char b[500];
  int parent, child;

  load_module(parent_name);
  load_module(child_name);

  scheduler_names(b, 500);
  parent = find_scheduler(b,parent_name);
  child = find_scheduler(b,child_name);

  bossa_connect(parent,child,child_property_count,child_properties);
}

void mount_root(char *new_root_name,
		int child_property_count,
		int *child_properties) {
  char b[500];
  int new_root, child;

  load_module(new_root_name);

  scheduler_names(b, 500);
  new_root = find_scheduler(b,new_root_name);

  bossa_connect_root(new_root,child_property_count,child_properties);
}

void unmount(char *tgt_name) {
  char b[500];
  int tgt;

  load_module(tgt_name);

  scheduler_names(b, 500);
  tgt = find_scheduler(b,tgt_name);

  bossa_disconnect(tgt);
}
